from ctypes import c_void_p, c_int, c_double, c_float, c_bool
from numpy cimport ndarray
cimport numpy as np
import numpy as np

ctypedef int idx_t
ctypedef unsigned long uint64_t

cdef extern from "partition.h":
    void dynamico_partition_graph(int mpi_rank, int mpi_size,
                     idx_t vtxdist[], idx_t xadj_loc[], idx_t adjncy_loc[],
                     idx_t nparts, idx_t part[])
    void dynamico_morton_encode(int n, const int *x, const int *y, const int *z, long *m)


#------------------------ Mesh partitioning ------------------------

# Helper functions and interface to ParMETIS
# loc_stencil returns the start/end indices (vtxdist) expected by ParMETIS
# i.e. index[start:end] with start=vtxdist[cell], end=vtxdist[cell+1] lists the edges of cell 'cell'
# list_stencil converts an adjacency graph from array format index[num_cells, MAX_EDGES] to compressed format

def list_stencil(degree, stencil, cond=lambda x:True):
    for i in range(degree.size):
        for j in range(degree[i]):
            s=stencil[i,j]
            if cond(s): yield stencil[i,j]

def loc_stencil(degree, stencil):
    loc=0
    for i in range(degree.size):
        yield loc
        loc=loc+degree[i]
    yield loc
        
def partition_mesh(degree, stencil, nparts): 
    # arguments : PArray1D and PArray2D describing mesh, number of desired partitions
    cdef int[::1] adjncy_loc, xadj_loc, vtxdist, owner
    dim_cell, degree, stencil = degree.dim, degree.data, stencil.data
    comm, vtxdist, idx_start, idx_end = dim_cell.comm, dim_cell.vtxdist, dim_cell.start, dim_cell.end
    adjncy_loc, xadj_loc = [np.asarray(list(x), dtype=np.int32)
                            for x in (list_stencil(degree, stencil), loc_stencil(degree, stencil)) ]
    owner = np.zeros(idx_end-idx_start, dtype=np.int32)
    cdef int mpi_rank = comm.Get_rank()
    cdef int mpi_size = comm.Get_size()
    dynamico_partition_graph(mpi_rank, mpi_size, &vtxdist[0], &xadj_loc[0], &adjncy_loc[0], nparts, &owner[0])
    return owner

def partition_graph(comm, int[::1] vtxdist, int[::1] xadj_loc, int[::1] adjncy_loc, int[::1] owner, int nparts):
    cdef int mpi_rank = comm.Get_rank()
    cdef int mpi_size = comm.Get_size()
    dynamico_partition_graph(mpi_rank, mpi_size, &vtxdist[0], &xadj_loc[0], &adjncy_loc[0], nparts, &owner[0])
    return owner

#-------------------------------- Space-filling (Morton) ordering ------------------------#

def toint(x): return np.int32((x+1.)*65536)

def morton_index(x,y,z):
    cdef int nn
    cdef int[::1] xx,yy,zz
    cdef long[::1] idx_
    nn, xx, yy, zz = x.size, toint(x), toint(y), toint(z)
    idx=np.zeros(nn, dtype=np.int64)
    idx_=idx
    dynamico_morton_encode(nn, &xx[0],&yy[0],&zz[0], &idx_[0])
    return idx
